'''
Created on Jun 15, 2015

@author: jldupont
'''

class MetaAction(type):
    
    actions = []
    
    def __new__(cls, future_class_name, future_class_parents, future_class_attr):
        newclass=super(MetaAction, cls).__new__(cls, future_class_name, future_class_parents, future_class_attr)
        
        if newclass.__name__ != 'Action':
            cls.actions.append(newclass)
            newclass.name = newclass.__name__
    
        return newclass
    

class Action(object):
    '''
    Base class for Actions
    '''
    
    __metaclass__ = MetaAction

    def __repr__(self, *args, **kwargs):
        return self.__class__.__name__



class Read(Action):
    '''
    'Read' Action on a resource
    '''

class Update(Action):
    '''
    'Update' Action on a resource
    '''
   
class Create(Action):
    '''
    'Create' Action for a resource
    '''
    
class Delete(Action):
    '''
    'Delete' Action on a resource
    '''
    
class List(Action):
    '''
    'List' Action for a collection of resource
    '''
    
class Search(Action):
    '''
    'Search' Action on a collection of resource
    '''

    



'''
Access Control

Created on Jun 14, 2015

@author: jldupont
'''

from .user import UserRBAC
from .permission import Permission, PermissionError

def ensure(user, permission):
    '''
    Ensures that 'user' has 'permission'
    
    Go through all the 'roles' assigned to a user type and
     confirm that there is at least 1 of these roles that
     list the required 'permission'. Else, raise 'PermissionError'.

    @param user: subclass of BaseUserRBAC
    @param permission: subclass of Permission  
    
    @return: None
    @raise PermissionError 
    '''
    assert isinstance(user, UserRBAC) or issubclass(user, UserRBAC), "The parameter 'user' must be an instance of 'BaseUserRBAC'"
    assert isinstance(permission, Permission), "The parameter 'permission' must be an instance of 'Permission'"
    
    for role in user.roles:
        if role.has_all_permissions:
            return
        
        if has_permission(role, permission):
            return
        
    raise PermissionError('Expecting %s on User %s' % (permission, user))
    

def has_permission(role, permission):
    '''
    Verifies that 'role' has 'permission'
    
    @param role: subclass of Role
    @param permission: subclass of Permission  
    
    @return True | False
    '''
    
    for _permission in role.permissions:
        
        if _permission == permission:
            return True

    return False

def can_assign(user_source, target_role):
    '''
    Can the 'user_source' assign 'target_role' to another user ?
    
    For 'user_source' to be able to assign 'target_role' to another user,
     'user_source' must have all the permissions assigned to 'target_role' *and* more.
    
    @return True | False
    '''
    
    target_permission_set = set(target_role.permissions)
    
    src_permissions_set = set()
    
    for role in user_source.roles:
        
        ##
        ## It only takes 1 admin (or equivalent) role
        ##
        if role.has_all_permissions:
            return True
        
        src_permissions_set.update(role.permissions)

    ##
    ## Must be a strict subset (i.e. not equal)
    ##        
    return target_permission_set < src_permissions_set



'''

Exporting to JSON model for integration with client side javascript
 e.g. https://github.com/jldupont/jsrbac

Created on Jul 4, 2015

@author: jldupont
'''
import json
import role

__all__ = [ 'export' ]

def export(tojson=True):
    '''
    Export to JSON representation compatible with 
     https://github.com/jldupont/jsrbac
     
    @param tojson: if True, returns JSON string representation
                   if False, returns a python dict ready for json.dumps
    '''
    
    roles = role.get_roles()
    
    edict = {}
    
    for one_role in roles:
        
        edict[one_role.name.lower()] = _export_role(one_role)
        
    if tojson:
        return json.dumps( edict )
    
    return edict

def _export_role(r):
    '''
    Exports a role
    '''
    def fn(x):
        return repr(x).lower()
    
    perms = map(fn, r.permissions)
    
    return {
             'a': r.has_all_permissions
            ,'p': perms
            }


'''
Created on Jun 14, 2015

@author: jldupont
'''

from role import get_role_from_name, get_role_from_name_or_class, roles_class_to_name_list


class UserRBAC(object):
    '''
    The base definition of a User
    '''

    def __init__(self, roles_name_or_classes = []):
        
        self.init_roles_from_names_or_classes(roles_name_or_classes)
        
        
    def init_roles_from_names_or_classes(self, names_or_classes):
        '''
        Initialize the user roles from a list of role names
         or a list of classes
        
        @raise RoleUnknownError
        '''
        self.roles = []        
        for name_or_class in names_or_classes:
            
            role_class = get_role_from_name_or_class(name_or_class)
            self.roles.append( role_class )
            
    def serialize_roles(self):
        '''
        Return a structure representing the roles 
         of the user. This structure can be persisted easily through
         a representation such as JSON.
         
        @return [string]
        '''
        return roles_class_to_name_list(self.roles)
             

    def init_roles_from_names(self, roles_name = []):
        '''
        Initialize the user roles from a list of role names
        
        @raise RoleUnknownError
        '''
        assert isinstance(roles_name, list)
        
        self.roles = []
         
        for role_name in roles_name:
            role_class = get_role_from_name(role_name)
            self.roles.append(role_class)
    
    def __repr__(self):
        return "User Roles: %s" % repr(self.roles)



'''
Created on Jun 15, 2015

@author: jldupont
'''
import unittest

from pyrbac import * #@UnusedWildImport



#
# ==================================================================================
#

class Domain(Resource):
    '''
    A test resource
    '''

class TestResource(Resource):
    '''
    A test resource
    '''



class Manager(Role):
    '''
    A test role
    '''
    permissions = define_permissions([
                                        (Domain, Create)
                                       ,(Domain, Read)
                                       ,(Domain, Update)
                                      ])

class Tester(Role):
    '''
    A test role
    '''
    permissions = define_permissions([
                                        (TestResource, List)
                                       ,(TestResource, Read)
                                       ,(TestResource, Update)
                                      ])


class Director(Role):
    '''
    A test role
    '''
    permissions = define_permissions([
                                        (Domain, Create)
                                       ,(Domain, Read)
                                       ,(Domain, Update)
                                       ,(Domain, Delete)
                                      ])


#
# ===================================================================================
#




class Test2(unittest.TestCase):
    '''
    Hypothetical roles tests
    '''


    def testPermission1a(self):
        '''
        Permission OK 2 - with user class instance
        '''
        user1 = UserRBAC([Manager])
                
        ensure(user1, Permission(Domain, Create))
        ensure(user1, Permission(Domain, Read))
        ensure(user1, Permission(Domain, Update))


    def testPermission1b(self):
        '''
        Another user class
        '''
        user2 = UserRBAC([Director])
        
        ensure(user2, Permission(Domain, Create))
        ensure(user2, Permission(Domain, Read))
        ensure(user2, Permission(Domain, Update))
        ensure(user2, Permission(Domain, Delete))  

    def testPermission1c(self):
        '''
        Multiple roles per user class
        '''
        user3 = UserRBAC([Manager, Tester])
                
        ensure(user3, Permission(Domain, Create))
        ensure(user3, Permission(Domain, Read))
        ensure(user3, Permission(Domain, Update))
        
        ensure(user3, Permission(TestResource, List))
        ensure(user3, Permission(TestResource, Read))
        ensure(user3, Permission(TestResource, Update))

    def testPermission2(self):
        '''
        Multiple roles per user class
        '''
        user3 = UserRBAC(['manager', 'tester'])
                
        ensure(user3, Permission(Domain, Create))
        ensure(user3, Permission(Domain, Read))
        ensure(user3, Permission(Domain, Update))
        
        ensure(user3, Permission(TestResource, List))
        ensure(user3, Permission(TestResource, Read))
        ensure(user3, Permission(TestResource, Update))

    def testPermission3(self):
        '''
        Missing permission
        '''        
        user1 = UserRBAC([Manager])
        
        with self.assertRaises(PermissionError):
            ensure(user1, Permission(TestResource, Create))

    def testAssignment1(self):
        '''
        A Director attempts to assign a role of 'Manager'
        '''
        
        user1 = UserRBAC([Director])
        
        self.assertTrue( can_assign(user1, Manager) )

    def testAssignment2(self):
        '''
        A Manager attempts to assign a role of 'Director'
        '''
        
        user1 = UserRBAC([Manager])
        
        self.assertFalse( can_assign(user1, Director) )

    def testAssignment3(self):
        '''
        A Manager attempts to assign a role of 'Manager'
        '''
        
        user1 = UserRBAC([Manager])
        
        self.assertFalse( can_assign(user1, Manager) )



'''
Created on Jun 14, 2015

@author: jldupont
'''
import unittest

from pyrbac.permission import * #@UnusedWildImport
import pyrbac.role as mrole

#
# ==================================================================================
#

class DummyClass(object): pass



class Test1(unittest.TestCase):
    '''
    Very basic tests
    '''

    def testRoleAdmin1(self):
        '''
        Existence
        '''
        self.assertTrue(mrole.Admin in mrole.get_roles(), 'Expecting Admin role')
        
    def testRoleAdmin2(self):
        '''
        Name property
        '''
        self.assertEqual((mrole.Admin).name, 'Admin', 'Expecting Admin class to have string name Admin')


    def testPermissionCreation1(self):
        '''
        Invalid creation 1
        '''
        with self.assertRaises(TypeError):
            Permission(None, None)


    def testPermissionCreation2(self):
        '''
        Invalid creation 2
        '''        
        with self.assertRaises(AssertionError):
            Permission(DummyClass, DummyClass)
            
    def testRoleSerialization(self):
        
        roles_list = [mrole.Admin, mrole.Guest]
        
        names_list = mrole.roles_class_to_name_list(roles_list)
        
        self.assertTrue('Admin' in names_list, 'Expected Admin in list, got %s' % repr(names_list))
        self.assertTrue('Guest' in names_list, 'Expected Guest in list, got %s' % repr(names_list))
        
            
    def testRoleDeserialization(self):
        
        names_list = ['Admin', 'Guest']
        
        roles_list = mrole.roles_class_from_name_list(names_list)
        self.assertTrue(mrole.Admin in roles_list, 'Expected role Admin in result list')
        self.assertTrue(mrole.Guest in roles_list, 'Expected role Guest in result list')



if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Jun 14, 2015

@author: jldupont
'''

from .permission import Permission


class RoleUnknownError(Exception): pass


class MetaRole(type):
    '''
    Metaclass for Role
    
    Collects all subclasses of Role
    '''
    
    roles = {}
    
    def __new__(cls, future_class_name, future_class_parents, future_class_attr):
        newclass=super(MetaRole, cls).__new__(cls, future_class_name, future_class_parents, future_class_attr)
        
        if newclass.__name__ != 'Role':
            
            name = newclass.__name__
            newclass.name = name
            cls.roles[name.lower()] = newclass
            newclass._validate()
    
        return newclass
    

def get_roles():
    '''
    Return all Role subclasses
    '''
    return MetaRole.roles.values()


def get_role_from_name(name):
    '''
    Return a Role class given the name of a role
    
    @raise RoleUnknownError 
    '''
    try:
        return MetaRole.roles[name.lower()]
    except:
        RoleUnknownError(str(name))


def get_role_from_name_or_class(name_or_class):
    '''
    Return a Role class based on either a role name
     or pass-through the Role class
     
    @return Role class
    @raise RoleUnknownError
    '''
    if isinstance(name_or_class, basestring):
        role_class = get_role_from_name(name_or_class)
        return role_class
    
    elif issubclass(name_or_class, Role):
        return name_or_class
    
    raise RoleUnknownError(str())
    
def roles_class_from_name_list(names_list):
    '''
    Construct [Role] from [string]
    
    @return [Role]
    @raise RoleUnknownError
    '''
    assert isinstance(names_list, list), 'Expecting [string], got %s' % repr(names_list)
    
    roles = []
    
    for name in names_list:
        roles.append( get_role_from_name(name) )
    
    return roles
    
def roles_class_to_name_list(roles_class):
    '''
    Return a list of role names given a list of role classes
    
    @return [string]
    @raise RoleUnknownError 
    '''
    assert isinstance(roles_class, list), 'Expecting list of Role class'
    
    roles_name = []
    for role_class in roles_class:
        assert issubclass(role_class, Role), 'expecting Role subsclass'
        roles_name.append( role_class.name )
        
    return roles_name

class Role(object):
    '''
    Base class for all Role
    '''

    __metaclass__ = MetaRole
    
    permissions = []
    has_all_permissions = False
    
    @classmethod
    def _validate(cls):
        '''
        Validate the permissions list
        '''
        for permission in cls.permissions:
            assert isinstance(permission, Permission), "got: %s" % repr(permission)
            
#
# ========================================================================= The basic roles
#
    
class Admin(Role):
    '''
    The most privileged role
    '''

    has_all_permissions = True
    

class Guest(Role):
    '''
    The least privileged role
    '''




'''
Created on Jun 14, 2015

@author: jldupont
'''

""" Can't use a metaclass with AppEngine ndb.Model 

class MetaResource(type):
    '''
    Metaclass for Resource
    
    Collects all subsclasses of Role
    '''
    
    resources = []
    
    def __new__(cls, future_class_name, future_class_parents, future_class_attr):
        newclass=super(MetaResource, cls).__new__(cls, future_class_name, future_class_parents, future_class_attr)
        
        if newclass.__name__ != 'Resource':
            cls.resources.append(newclass)
            newclass.name = newclass.__name__
    
        return newclass


def get_resources():
    '''
    Returns the list of resource classes 
    '''
    return MetaResource.resources
"""
#
#
#

class Resource(object):
    '''
    Base class for a Resource
    '''
    
    #__metaclass__ = MetaResource
    


'''
Created on Jun 14, 2015

@author: jldupont
'''

from .resource import Resource
from .action import Action

class PermissionError(Exception):
    '''
    The exception class for Permission access control
    '''


class Permission(object):
    '''
    Base class for permissions
    '''

    def __init__(self, resource, action):

        assert issubclass(resource, Resource), "Expecting subclass of Resource"
        assert issubclass(action,   Action),   "Expecting subclass of Action"

        self.for_resource = resource
        self.for_action   = action
    
    def __repr__(self, *args, **kwargs):
        return "%s:%s" % (self.for_resource.__name__, self.for_action.__name__)

    def __eq__(self, other):
        return self.for_resource == other.for_resource and self.for_action == other.for_action

    def __hash__(self):
        '''
        Useful in set() operations
        '''
        return hash(self.__repr__())

    
def define_permissions(tuples_resource_action):
    '''
    Helper function for defining Permissions
    
    @return [Permission] 
    '''
    permissions = []
    
    for resource, action in tuples_resource_action:
        assert issubclass(resource, Resource), "expecting Resource in tuple (Resource, Action)"
        assert issubclass(action,   Action),   "expecting Action in tuple (Resource, Action)"
        
        permissions.append(Permission(resource, action))
        
    return permissions


'''
Created on Jun 15, 2015

@author: jldupont
'''
import unittest

from pyrbac.user import UserRBAC
from pyrbac.action import * #@UnusedWildImport
from pyrbac.role import Admin, Guest


#
# ==================================================================================
#
class User1(UserRBAC):
    '''
    A test user
    '''
    def __init__(self, roles_name):
        super(User1, self).__init__(roles_name)


#
# ===================================================================================
#




class Test3(unittest.TestCase):
    '''
    User class related tests
    '''
    
    roles1 = ['Admin']
    
    def setUp(self):
        self.user1 = User1(self.roles1)

    def testUser1(self):
        
        assert isinstance(self.user1, UserRBAC)
        
    def testUser2(self):
        
        user = User1([Admin, Guest])
        
        roles = user.serialize_roles()
        
        self.assertTrue('Admin' in roles, 'Expecting to find string Admin')
        self.assertTrue('Guest' in roles, 'Expecting to find string Guest')




'''
Created on July 4, 2015

@author: jldupont
'''
import unittest

from pyrbac import export


#
# ===================================================================================
#
class Test4(unittest.TestCase):
    '''
    Export
    
    The relevant Role classes are defined in tests_2
    '''
    def testExport1(self):
        
        edict = export(tojson=False)
        
        self.assertEqual(edict['admin']['a'], True, "Expecting 'all permission' from Admin")
        self.assertEqual('domain:create' in edict['manager']['p'], True, "Expecting 'Domain:Create' from Manager")


#!/usr/bin/env python
"""
    Jean-Lou Dupont's Python RBAC (Role Based Access Control)
    
    Created on 2015-06-14
    @author: jldupont
"""
__author__  ="Jean-Lou Dupont"
__version__ ="0.4"

DESC="""
RBAC (Role Based Access Control) library loosely based on NIST model (see http://csrc.nist.gov/groups/SNS/rbac/ ).
"""


from distutils.core import setup
from setuptools import find_packages


setup(name=         'pyrbac',
      version=      __version__,
      description=  'RBAC (Role Based Access Control) library',
      author=       __author__,
      author_email= 'jl@jldupont.com',
      url=          'https://github.com/jldupont/pyrbac',
      package_dir=  {'': "src",},
      packages=     find_packages("src"),
      scripts=      [
                     ]
      ,zip_safe=True
      ,long_description=DESC
      ,install_requires=[
                         ]
      ,test_suite="tests"
      ,use_2to3 = True
      )

#############################################

f=open("latest", "w")
f.write(str(__version__)+"\n")
f.close()

